#include "dictionary_t.h"

//---------------------------------------------------
// Constructors/Destructor 
//---------------------------------------------------

//---------------------------------------------------
// Default constructor
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t>::Dictionary_T()
  : myEntries(NULL)
{
}


//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t>::Dictionary_T(DictNode_T<Key_t, Datum_t> *n)
  : myEntries(n)
{
  if (myEntries != NULL) {
    myEntries->IncrRef();
  }
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t>::Dictionary_T(Key_t k, Datum_t v, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r)
  : myEntries(new DictNode_T<Key_t, Datum_t>(k,v,l,r))
{
  myEntries->IncrRef();
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t>::Dictionary_T(Key_t k, Datum_t v)
  : myEntries(new DictNode_T<Key_t, Datum_t>(k,v))
{
  myEntries->IncrRef();
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t>::Dictionary_T(const Dictionary_T<Key_t, Datum_t> &d)
{
  myEntries = d.myEntries;
  if (myEntries != NULL) {
    myEntries->IncrRef();
  }
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t>::~Dictionary_T()
{
  if (myEntries != NULL) {
    myEntries->DecrRef();
  }
}

//---------------------------------------------------
// Overloaded assignment
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> &Dictionary_T<Key_t, Datum_t>::operator= (const Dictionary_T<Key_t, Datum_t> &d)
{
  if (this != &d)      // don't assign to self!
  {
    DictNode_T<Key_t, Datum_t> *temp = myEntries;
    myEntries = d.myEntries;
    if (myEntries != NULL) {
      myEntries->IncrRef();
    }
    if (temp != NULL) {
      temp->DecrRef();
    }
  }
  return *this;    	
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> Dictionary_T<Key_t, Datum_t>::Insert(Key_t k, Datum_t v)
{
  if (myEntries == NULL)
    return Dictionary_T<Key_t, Datum_t>(k,v);
  int temp = keycompare(k, Key());
  if (temp < 0)
    return Bal(Key(), Datum(), Left().Insert(k,v), Right());
  else if (temp == 0)
    return Bal(k,v,Left(),Right());
  else
    return Bal(Key(), Datum(), Left(), Right().Insert(k,v));
}



//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> Dictionary_T<Key_t, Datum_t>::Delete(Key_t k)
{
  if (myEntries == NULL) return NULL;
  int temp = keycompare(k, Key());
  if (temp < 0)
    return Bal(Key(), Datum(), Left().Delete(k), Right());
  else if (temp == 0)
  {
    if ((myEntries->left).NoEntries()) return myEntries->right;
    if ((myEntries->right).NoEntries()) return myEntries->left;
    Dictionary_T<Key_t, Datum_t> rmost = Left().RightMost();
    return Bal(rmost.Key(), rmost.Datum(), Left().Delete(rmost.Key()), Right());
  }
  else
    return Bal(Key(), Datum(), Left(), Right().Delete(k));
}


//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
STATUS Dictionary_T<Key_t, Datum_t>::Lookup(Key_t k, Datum_t &v)
{
  if (myEntries == NULL)
    return NOT_FOUND;
  int temp = keycompare(k, Key());
  if (temp < 0) {
    if (Left().NoEntries())
      return NOT_FOUND;
    else
      return Left().Lookup(k,v);
  }
  else if (temp == 0)
  {
    v = myEntries->datum;
    return FOUND;
  }
  else {
    if (Right().NoEntries())
      return NOT_FOUND;
    else
      return Right().Lookup(k,v);
  }
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> Dictionary_T<Key_t, Datum_t>::RightMost()
{
  if (Right().NoEntries())
    return *this;
  else
    return (myEntries->right).RightMost();
}


//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
bool Dictionary_T<Key_t, Datum_t>::NoEntries()
{
  return (myEntries == NULL);
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Key_t Dictionary_T<Key_t, Datum_t>::Key()
{
	assert(myEntries!=NULL);
	return myEntries->key;
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Datum_t Dictionary_T<Key_t, Datum_t>::Datum()
{
	assert(myEntries!=NULL);
    return myEntries->datum;
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> Dictionary_T<Key_t, Datum_t>::Left()
{
  if (myEntries == NULL)
    return Dictionary_T<Key_t, Datum_t>(NULL);
  else
    return myEntries->left;
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> Dictionary_T<Key_t, Datum_t>::Right()
{
  if (myEntries == NULL)
    return Dictionary_T<Key_t, Datum_t>(NULL);
  else
    return myEntries->right;
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
height_t Dictionary_T<Key_t, Datum_t>::Height()
{
  if (myEntries == NULL)
    return 0;
  else
    return myEntries->height;
}


// Given two Dictionaries, l and r, whose heights are within two,
// create a balanced Dictionary_T<Key_t, Datum_t> from k, d, l, and r.
//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
Dictionary_T<Key_t, Datum_t> Dictionary_T<Key_t, Datum_t>::Bal(Key_t k, Datum_t d, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r)
{
  if (WithinOne(l.Height(), r.Height()))
    return Dictionary_T<Key_t, Datum_t>(k,d,l,r);
  else if (l.Height() > r.Height())
  { // l.Height() >= 2
    if (l.Left().Height() >= l.Right().Height()) {
      // cout << "Single rotation at " << k << ": " << d << endl;
      return Dictionary_T<Key_t, Datum_t>(l.Key(), l.Datum(), l.Left(), new DictNode_T<Key_t, Datum_t>(k,d,l.Right(),r));
    }
    else {
      // cout << "Double rotation at " << k << ": " << d << endl;
      return Dictionary_T<Key_t, Datum_t>(l.Right().Key(),
                        l.Right().Datum(),
                        Dictionary_T<Key_t, Datum_t>(l.Key(), l.Datum(), l.Left(), l.Right().Left()),
                        Dictionary_T<Key_t, Datum_t>(k, d, l.Right().Right(), r)
                       );
    }
  }
  else
  { // r.Height() >= 2
    if (r.Right().Height() >= r.Left().Height()) {
      // cout << "Single rotation at " << k << ": " << d << endl;
      return Dictionary_T<Key_t, Datum_t>(r.Key(), r.Datum(), new DictNode_T<Key_t, Datum_t>(k,d,l,r.Left()), r.Right());
    }
    else {
      // cout << "Double rotation at " << k << ": " << d << endl;
      return Dictionary_T<Key_t, Datum_t>(r.Left().Key(),
                        r.Left().Datum(),
                        Dictionary_T<Key_t, Datum_t>(k,d,l,r.Left().Left()),
                        Dictionary_T<Key_t, Datum_t>(r.Key(), r.Datum(), r.Left().Right(), r.Right())
                       );
    }
  }
}
//---------------------------------------------------
// Iterators
//---------------------------------------------------
//---------------------------------------------------
// First 

/*
		
	private:
		bool rightDone;
		iterator *parent;
		Dictionary_T<Key_t, Datum_t> *myEntry;
*/

//---------------------------------------------------
template <class Key_t, class Datum_t>
typename Dictionary_T<Key_t, Datum_t>::iterator *Dictionary_T<Key_t, Datum_t>::First() {
	if(!myEntries)
		return NULL;
	//
	typename Dictionary_T<Key_t, Datum_t>::iterator *iter;
	typename Dictionary_T<Key_t, Datum_t>::iterator *prevEntry;
	prevEntry=NULL;
	DictNode_T<Key_t, Datum_t> *curNode=myEntries;
	while(curNode) {
		iter=new typename Dictionary_T<Key_t, Datum_t>::iterator();
		iter->parent=prevEntry;
		iter->myEntry=curNode;
		iter->rightDone=false;
		prevEntry=iter;
		curNode=curNode->left.myEntries;
	}
	return iter;
}


//---------------------------------------------------
// Print
//---------------------------------------------------
template <class Key_t, class Datum_t>
void Dictionary_T<Key_t, Datum_t>::Print(ostream & out, unsigned int depth) const
{
  if (myEntries == NULL) {
    //    out << "NULL" << endl;
    out << "NULL";
  }
  else {
    if (!(myEntries->left).NoEntries())
      (myEntries->left).Print(out, depth+1);

    //    for (unsigned int i = 0; i < depth; i++) out << "  ";
    //   out << "height: " << (myEntries->height) << ", key: " << (myEntries->key) << ", datum: " << (myEntries->datum) << ", refCount: " << myEntries->refCount << endl;
    out <<  "(" << (myEntries->key) << "," << (myEntries->datum) << ")";

   if (!(myEntries->right).NoEntries())
     (myEntries->right).Print(out, depth+1);
  }
}

static int WithinOne(unsigned int i, unsigned int j)
{
  return(i==j || i+1==j || j+1==i);
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
ostream& operator<< (ostream & out, const Dictionary_T<Key_t, Datum_t> &d)
{
  d.Print(out, 0);
  return(out);
}


//********************************************************************
// DictNode_T<Key_t, Datum_t>
//********************************************************************
//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
DictNode_T<Key_t, Datum_t>::DictNode_T(Key_t k, Datum_t d, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r, height_t h)
  : key(k), datum(d), left(l), right(r), height(h), refCount(0)
{
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
DictNode_T<Key_t, Datum_t>::DictNode_T(Key_t k, Datum_t d, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r)
  : key(k), datum(d), left(l), right(r), refCount(0)
{
  height_t l_height = l.Height();
  height_t r_height = r.Height();
  height = 1 + ( (l_height >= r_height) ? l_height : r_height );
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
DictNode_T<Key_t, Datum_t>::DictNode_T(Key_t k, Datum_t d)
  : key(k), datum(d), left(Dictionary_T<Key_t, Datum_t>()), right(Dictionary_T<Key_t, Datum_t>()), height(1), refCount(0)
{
}


//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
DictNode_T<Key_t, Datum_t>::~DictNode_T()
{
  // Destructors for left and right called automatically,
  // which will decrement the reference counts for left-myEntries
  // and right->myEntries (assuming they are non-NULL).
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
void DictNode_T<Key_t, Datum_t>::IncrRef()
{
  if (refCount < maxRefCount) {
    refCount++;
  }
}

//---------------------------------------------------
// 
//---------------------------------------------------
template <class Key_t, class Datum_t>
void DictNode_T<Key_t, Datum_t>::DecrRef()
{
  if (refCount < maxRefCount) {
    refCount -= 1;
    if (refCount == 0) {
      delete this;
    }
  }
}

//------------------------------------
